/**
 * eobjects.org MetaModel
 * Copyright (C) 2010 eobjects.org
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution; if not, write to:
 * Free Software Foundation, Inc.
 * 51 Franklin Street, Fifth Floor
 * Boston, MA  02110-1301  USA
 */
package org.eobjects.metamodel.query.builder;

import java.util.Collection;
import java.util.Date;

import org.eobjects.metamodel.query.FilterItem;
import org.eobjects.metamodel.query.OperatorType;
import org.eobjects.metamodel.query.SelectItem;
import org.eobjects.metamodel.schema.Column;

public abstract class AbstractFilterBuilder<B> implements FilterBuilder<B> {

    private final SelectItem _selectItem;

    public AbstractFilterBuilder(SelectItem selectItem) {
        this._selectItem = selectItem;
    }

    protected abstract B applyFilter(FilterItem filter);

    @Override
    public B in(Collection<?> values) {
        return applyFilter(new FilterItem(_selectItem, OperatorType.IN, values));
    }

    @Override
    public B in(Number... numbers) {
        return applyFilter(new FilterItem(_selectItem, OperatorType.IN, numbers));
    }

    @Override
    public B in(String... strings) {
        return applyFilter(new FilterItem(_selectItem, OperatorType.IN, strings));
    }

    @Override
    public B isNull() {
        return applyFilter(new FilterItem(_selectItem, OperatorType.EQUALS_TO, null));
    }

    @Override
    public B isNotNull() {
        return applyFilter(new FilterItem(_selectItem, OperatorType.DIFFERENT_FROM, null));
    }

    @Override
    public B equals(Column column) {
        if (column == null) {
            throw new IllegalArgumentException("column cannot be null");
        }
        return applyFilter(new FilterItem(_selectItem, OperatorType.EQUALS_TO, new SelectItem(column)));
    }

    @Override
    public B equals(Date date) {
        if (date == null) {
            throw new IllegalArgumentException("date cannot be null");
        }
        return applyFilter(new FilterItem(_selectItem, OperatorType.EQUALS_TO, date));
    }

    @Override
    public B equals(Number number) {
        if (number == null) {
            throw new IllegalArgumentException("number cannot be null");
        }
        return applyFilter(new FilterItem(_selectItem, OperatorType.EQUALS_TO, number));
    }

    @Override
    public B equals(String string) {
        if (string == null) {
            throw new IllegalArgumentException("string cannot be null");
        }
        return applyFilter(new FilterItem(_selectItem, OperatorType.EQUALS_TO, string));
    }

    @Override
    public B equals(Boolean bool) {
        if (bool == null) {
            throw new IllegalArgumentException("bool cannot be null");
        }
        return applyFilter(new FilterItem(_selectItem, OperatorType.EQUALS_TO, bool));
    }

    @Override
    public B differentFrom(Column column) {
        if (column == null) {
            throw new IllegalArgumentException("column cannot be null");
        }
        return applyFilter(new FilterItem(_selectItem, OperatorType.DIFFERENT_FROM, new SelectItem(column)));
    }

    @Override
    public B differentFrom(Date date) {
        if (date == null) {
            throw new IllegalArgumentException("date cannot be null");
        }
        return applyFilter(new FilterItem(_selectItem, OperatorType.DIFFERENT_FROM, date));
    }

    @Override
    public B differentFrom(Number number) {
        if (number == null) {
            throw new IllegalArgumentException("number cannot be null");
        }
        return applyFilter(new FilterItem(_selectItem, OperatorType.DIFFERENT_FROM, number));
    }

    @Override
    public B differentFrom(String string) {
        if (string == null) {
            throw new IllegalArgumentException("string cannot be null");
        }
        return applyFilter(new FilterItem(_selectItem, OperatorType.DIFFERENT_FROM, string));
    }

    @Override
    public B differentFrom(Boolean bool) {
        if (bool == null) {
            throw new IllegalArgumentException("bool cannot be null");
        }
        return applyFilter(new FilterItem(_selectItem, OperatorType.DIFFERENT_FROM, bool));
    }

    @Deprecated
    @Override
    public B higherThan(Column arg) {
        return greaterThan(arg);
    }

    @Override
    public B greaterThan(Column column) {
        if (column == null) {
            throw new IllegalArgumentException("column cannot be null");
        }
        return applyFilter(new FilterItem(_selectItem, OperatorType.GREATER_THAN, new SelectItem(column)));
    }

    @Deprecated
    @Override
    public B higherThan(Date arg) {
        return greaterThan(arg);
    }

    @Override
    public B greaterThan(Date date) {
        if (date == null) {
            throw new IllegalArgumentException("date cannot be null");
        }
        return applyFilter(new FilterItem(_selectItem, OperatorType.GREATER_THAN, date));
    }

    @Deprecated
    @Override
    public B higherThan(Number arg) {
        return greaterThan(arg);
    }

    @Override
    public B greaterThan(Number number) {
        if (number == null) {
            throw new IllegalArgumentException("number cannot be null");
        }
        return applyFilter(new FilterItem(_selectItem, OperatorType.GREATER_THAN, number));
    }

    @Deprecated
    @Override
    public B higherThan(String arg) {
        return greaterThan(arg);
    }

    @Override
    public B greaterThan(String string) {
        if (string == null) {
            throw new IllegalArgumentException("string cannot be null");
        }
        return applyFilter(new FilterItem(_selectItem, OperatorType.GREATER_THAN, string));
    }

    @Override
    public B lessThan(Column column) {
        if (column == null) {
            throw new IllegalArgumentException("column cannot be null");
        }
        return applyFilter(new FilterItem(_selectItem, OperatorType.LESS_THAN, new SelectItem(column)));
    }

    @Override
    public B lessThan(Date date) {
        if (date == null) {
            throw new IllegalArgumentException("date cannot be null");
        }
        return applyFilter(new FilterItem(_selectItem, OperatorType.LESS_THAN, date));
    }

    @Override
    public B lessThan(Number number) {
        if (number == null) {
            throw new IllegalArgumentException("number cannot be null");
        }
        return applyFilter(new FilterItem(_selectItem, OperatorType.LESS_THAN, number));
    }

    @Override
    public B lessThan(String string) {
        if (string == null) {
            throw new IllegalArgumentException("string cannot be null");
        }
        return applyFilter(new FilterItem(_selectItem, OperatorType.LESS_THAN, string));
    }

    @Override
    public B like(String string) {
        if (string == null) {
            throw new IllegalArgumentException("string cannot be null");
        }
        return applyFilter(new FilterItem(_selectItem, OperatorType.LIKE, string));
    }
}